/* 
 * File   : airsec_main.cpp
 * Version: 2010-02-07
 *
 * Copyright (C) 2010 Jochen Peters
 *
 * dev[at]emirp.net
 * http://www.emirp.net/
 * 
 * This file is part of Approach Control.
 *
 * Approach Control is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Approach Control is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Approach Control. If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <SDL.h>

#include "pathsprite.hpp"
#include "flyobj.hpp"


/* Defines */
#define WINDOWTITLE "Approach Control (Ver. 1.01) - www.emirp.net"
#define WINDOWICON "Approach Control"

#define MAXNUMSPRITES 20
#define MAXNUMTARGETS 3

#define GAMETICKDELAY 10 
#define SPEED_A 0.75
#define SPEED_B 0.85
#define SPEED_C 0.60
#define SPEED_D 0.45

#define DIGITWIDTH 10
#define DIGITHEIGHT 14
#define DIGITPOSY 5
#define SCOREPOSX 435
#define HIGHSCOREPOSX 590
#define PAUSEPOSY 90
#define PRESSKEYPOSY 425

#define HIGHSCOREFILE "highscore.txt"



/* Global variables */
Uint32 lastTick = 0;

SDL_Surface* screen;
SDL_Surface* digitSurf;
SDL_Surface* bgSurf;
SDL_Surface* txtPauseSurf;
SDL_Surface* txtPresskeySurf;

FlyObj* sprites[MAXNUMSPRITES];
unsigned int numsprites = 0;

Target* targets[MAXNUMTARGETS];
unsigned int numtargets = 0;

Marker* marker = NULL;

unsigned int score = 0;
unsigned int highscore = 0;



/* Call this instead of exit(), so we can clean up SDL: atexit() is evil. */
static void quit(int _rc) {
	SDL_Quit();
	exit(_rc);
}


void readHighscore() {

	FILE* f = fopen(HIGHSCOREFILE, "r");

	if(f != NULL) {
		int retval = fscanf(f, "%d", &highscore);
		fclose(f);
		if(retval == EOF) {
			highscore = 0;
		}
	}
	else {
		highscore = 0;
	}

} 

void writeHighscore() {

	FILE* f = fopen(HIGHSCOREFILE, "w");

	if(f != NULL) {
		fprintf(f, "%d", highscore);
		fclose(f);
	}

} 


void drawDigits(unsigned int _score, int _pos) {
	
	const int digitWidth = DIGITWIDTH;
	const int digitHeight = DIGITHEIGHT;

	SDL_Rect dpos;
	SDL_Rect spos;

	spos.x = _pos;
	spos.y = DIGITPOSY;
	spos.w = digitWidth;
	spos.h = digitHeight;
	dpos.x = 0;
	dpos.y = 0;
	dpos.w = digitWidth;
	dpos.h = digitHeight;

	unsigned int tmpScore = _score;
	unsigned int div = 1000;

	while(div >= 1) {
	
		dpos.x = ((int) tmpScore / div) * dpos.w;
		SDL_BlitSurface(digitSurf, &dpos, screen, &spos);
		spos.x += digitWidth;
		tmpScore = tmpScore % div;
		
		div /= 10;
	}

}


void drawScore() {
	drawDigits(score, SCOREPOSX);
	drawDigits(highscore, HIGHSCOREPOSX);
}



void redraw() {
	
	SDL_Rect tmpRect;

	/* Draw background */
	tmpRect.x = 0;
	tmpRect.y = 0;
	tmpRect.w = screen->w;
	tmpRect.h = screen->h;
	//SDL_FillRect(screen, &tmpRect, 0);
	SDL_BlitSurface(bgSurf, NULL, screen, &tmpRect);


	/* Draw score */
	drawScore();

	/* Draw targets (DEBUG) *
	for(int i=0; i<numtargets; i++) {
		targets[i]->draw(screen);
	}
	*/

	// START: OBSOLETE
	/* Draw color marker for targets
	tmpRect.w = 5;
	tmpRect.h = 5;
	tmpRect.x = 300;
	tmpRect.y = 195;
	SDL_FillRect(screen, &tmpRect, SDL_MapRGB(screen->format, 255, 0, 0));
	tmpRect.y = 230;
	SDL_FillRect(screen, &tmpRect, SDL_MapRGB(screen->format, 255, 0, 0));
	tmpRect.x = 395;
	tmpRect.y = 295;
	SDL_FillRect(screen, &tmpRect, SDL_MapRGB(screen->format, 0, 0, 255));
	tmpRect.x = 440;
	SDL_FillRect(screen, &tmpRect, SDL_MapRGB(screen->format, 0, 0, 255));
	tmpRect.y = 340;
	SDL_FillRect(screen, &tmpRect, SDL_MapRGB(screen->format, 0, 0, 255));
	tmpRect.x = 395;
	SDL_FillRect(screen, &tmpRect, SDL_MapRGB(screen->format, 0, 0, 255));
	*/
	// END: OBSOLETE

	/* Draw paths */
	for(int i=0; i<numsprites; i++) {
		sprites[i]->drawPathRGBA(screen, 0, 255, 0, 100);
	}
	/* Draw sprites */
	for(int i=0; i<numsprites; i++) {
		sprites[i]->draw(screen);
	}
	/* Draw marker */
	if(marker != NULL) {
		marker->draw(screen, SDL_MapRGB(screen->format, 255, 255, 255));
	}

	SDL_Flip(screen);

}


void gameoverScreen() {
	
	SDL_Rect tmpRect;

	tmpRect.x = (screen->w - txtPresskeySurf->w)/2;
	tmpRect.y = PRESSKEYPOSY;
	tmpRect.w = txtPresskeySurf->w;
	tmpRect.h = txtPresskeySurf->h;
	SDL_BlitSurface(txtPresskeySurf, NULL, screen, &tmpRect);

	SDL_UpdateRects(screen, 1, &tmpRect);

}


void pauseScreen() {
	
	SDL_Rect tmpRect;

	tmpRect.x = 0;
	tmpRect.y = 0;
	tmpRect.w = screen->w;
	tmpRect.h = screen->h;
	SDL_BlitSurface(bgSurf, NULL, screen, &tmpRect);

	drawScore();

	tmpRect.x = (screen->w - txtPauseSurf->w)/2;
	tmpRect.y = PAUSEPOSY;
	tmpRect.w = txtPauseSurf->w;
	tmpRect.h = txtPauseSurf->h;
	SDL_BlitSurface(txtPauseSurf, NULL, screen, &tmpRect);

	tmpRect.x = (screen->w - txtPresskeySurf->w)/2;
	tmpRect.y = PRESSKEYPOSY;
	tmpRect.w = txtPresskeySurf->w;
	tmpRect.h = txtPresskeySurf->h;
	SDL_BlitSurface(txtPresskeySurf, NULL, screen, &tmpRect);

	SDL_Flip(screen);

}


void resetSituation() {
	
	score = 0;

	for(int i=0; i<numsprites; i++) {
		sprites[i]->remove(screen, 0);
		sprites[i]->setNormal();
		sprites[i]->clearPath();
		sprites[i]->status = SPRITESTAT_IDLE;
		sprites[i]->setPos(0, 0);
	}

}


int clickOnSprite(int _x, int _y) {

	for(int i=0; i<numsprites; i++) {
		if(sprites[i]->collisionCheck(_x, _y)) {
			return i;
		}
	}

	return -1;
}


bool checkSpriteCollision(int* _sindex1, int* _sindex2) {

	for(int i=0; i<(numsprites-1); i++) {
		for(int j=(i+1); j<numsprites; j++) {
			if(sprites[i]->collisionCheckOutline(sprites[j])) {
				*_sindex1 = i;
				*_sindex2 = j;
				return true;
			}
		}
	}

	return false;
}


void doGameoverTasks() {
	
	gameoverScreen();
	
	if(score > highscore) {
		highscore = score;
		writeHighscore();
	}

	resetSituation();

}


bool doGameTickTasks() {

	bool gameover = false;

	/* Move sprites */
	for(int i=0; i<numsprites; i++) {
		sprites[i]->move(screen, 0);
	}
	
	/* Check for target collision */
	for(int i=0; i<numsprites; i++) {
		if(sprites[i]->checkTarget(targets, numtargets)) {
			sprites[i]->remove(screen, 0);
			score++;
		}
	}

	/* Check for sprite collision */
	int s1 = 0;
	int s2 = 0;
	if(checkSpriteCollision(&s1, &s2)) {
		
		sprites[s1]->status = SPRITESTAT_COLLISION;
		sprites[s2]->status = SPRITESTAT_COLLISION;
		sprites[s1]->setCrashed();
		sprites[s2]->setCrashed();
		sprites[s1]->draw(screen);
		sprites[s2]->draw(screen);

		redraw();

		gameover = true;
	}

	/* Start new (marked) sprites */
	if(marker != NULL) {
		if(marker->startSprite(screen)) {
			delete marker;
			marker = NULL;
		}
	}

	/* Redraw screen */
	redraw();


	/* Timing */
	if ((SDL_GetTicks() - lastTick) < GAMETICKDELAY) {
		SDL_Delay(GAMETICKDELAY - (SDL_GetTicks() - lastTick));
	}
	lastTick = SDL_GetTicks();


	/* Handle game over */
	if(gameover) {
		doGameoverTasks();
		return false;
	}
	else {
		return true;
	}

}


bool handleMouseInput(SDL_Event* _event) {

	int sindex = clickOnSprite(_event->button.x, _event->button.y);

	if(sindex >= 0) {
	
		Uint16* buffer16;
		Uint32 color;
		int last_x = -1;
		int last_y = -1;


		for(;;) {

			/* Get new coordinates */
			int x = _event->button.x;
			int y = _event->button.y;
						

			// START: OBSOLETE
			/* Draw and save path *
			if(SDL_LockSurface(screen) < 0 ) {
				fprintf(stderr, "Couldn't lock the display surface: %s\n", SDL_GetError());
				quit(2);
			}
		
			color = SDL_MapRGB(screen->format, 0, 255, 0);
			buffer16 = (Uint16*) screen->pixels;
		
			if(last_x == -1 || last_y == -1) { // first click: draw pixel
				sprites[sindex]->clearPath();
				buffer16[x + y*screen->pitch/screen->format->BytesPerPixel] = (Uint16) color;
			}
			else { // connecting: draw line
				lineRGBA(screen, last_x, last_y, x, y, 0, 255, 0, 100); // SDL-gfx (!)
			}
		
			SDL_UnlockSurface(screen);
			SDL_Flip(screen);
			*/
			// END: OBSOLETE


			/* Save path */
			if(last_x == -1 || last_y == -1) { // first click: clear old path
				sprites[sindex]->clearPath();
			}

			if(!sprites[sindex]->addToPath(x, y)) {
				last_x = -1;
				last_y = -1;
				break;
			}
			else {
				last_x = x;
				last_y = y;
			}
			
			SDL_PollEvent(_event);
			if(_event->type == SDL_MOUSEBUTTONUP) {
				last_x = -1;
				last_y = -1;
				break;
			}
	
			/* Move sprites, check for collisions, ... */
			if(!doGameTickTasks()) {
				return false;
			}
	
		}
	
		/* Check if path hits target */
		if(sprites[sindex]->checkPath(targets, numtargets)) {
			sprites[sindex]->setOnApproach();
			sprites[sindex]->status = SPRITESTAT_ONAPPROACH;
		}
		else {
			sprites[sindex]->setNormal();
			sprites[sindex]->status = SPRITESTAT_ONPATH;
		}
		sprites[sindex]->draw(screen);
	
	}

	return true;

}


void addNewSprite() {

	if(marker == NULL) {

		unsigned int dt = 3000 - 10*score;
		if(dt < MAXNUMSPRITES) {
			dt = MAXNUMSPRITES;
		}

		/* Calculate random border position */
		int rndsprite = rand() % dt;
		if(rndsprite < MAXNUMSPRITES && sprites[rndsprite]->status == SPRITESTAT_IDLE) {
	
			int rndpos = rand() % (2*screen->w + 2*screen->h);
			int posx = 0;
			int posy = 0;
			int velx = 0;
			int vely = 0;

			SDL_Rect secArea;
			secArea.w = 80;
			secArea.h = 80;
	
			if(rndpos < screen->w) { // top
				posx = rndpos - sprites[rndsprite]->getWidth();
				if(posx < 0) posx = 0;
				posy = 0;
				velx = 0;
				vely = 1;
				secArea.x = posx - secArea.w/2;
				if(secArea.x < 0) secArea.x = 0;
				secArea.y = 0;
			}
			else if(rndpos < screen->w + screen->h) { // left
				posx = 0;
				posy = rndpos - screen->w - sprites[rndsprite]->getHeight();
				if(posy < 0) posy = 0;
				velx = 1;
				vely = 0;
				secArea.x = 0;
				secArea.y = posy - secArea.h/2;
				if(secArea.y < 0) secArea.y = 0;
			}
			else if(rndpos < 2*screen->w + screen->h) { // bottom
				posx = rndpos - screen->w - screen->h - sprites[rndsprite]->getWidth();
				if(posx < 0) posx = 0;
				posy = screen->h - sprites[rndsprite]->getHeight();
				velx = 0;
				vely = -1;
				secArea.x = posx - secArea.w/2;
				if(secArea.x < 0) secArea.x = 0;
				secArea.y = screen->h - secArea.h;
			}
			else {
				posx = screen->w - sprites[rndsprite]->getWidth(); // right
				posy = rndpos - 2*screen->w - screen->h - sprites[rndsprite]->getHeight();
				if(posy < 0) posy = 0;
				velx = -1;
				vely = 0;
				secArea.x = screen->w - secArea.w;
				secArea.y = posy - secArea.h/2;
				if(secArea.y < 0) secArea.y = 0;
			}

			/* Check if some sprite is around */
			for(int i=0; i<numsprites; i++) {
				if(sprites[i]->collisionCheckOutline(&secArea)) {
					return; // leave function without placing a new sprite (!)
				}
			}
			
			/* Place marker for new sprite */
			sprites[rndsprite]->setPos(posx, posy);
			sprites[rndsprite]->setVelocity(velx, vely);
			sprites[rndsprite]->clearPath();
			sprites[rndsprite]->setNormal();
			sprites[rndsprite]->status = SPRITESTAT_FREERUNNING;
			marker = new Marker(sprites[rndsprite], 100);
	
		}
	
	}

}



int main(int argc, char *argv[]) {

	int i;
	Uint8 video_bpp = 16;
	Uint32 videoflags = SDL_SWSURFACE;
	SDL_Event event;

	/* Initialize random number generator */
	srand(time(NULL));

	/* Initialize SDL */
	if(SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
		return(1);
	}

	/* Set video mode */
	if((screen=SDL_SetVideoMode(640, 480, video_bpp, videoflags)) == NULL) {
		fprintf(stderr, "Couldn't set video mode: %s\n", SDL_GetError());
		quit(2);
	}

	SDL_WM_SetCaption(WINDOWTITLE, WINDOWICON);

	/* Load bitmaps */
	SDL_Surface* tmpSurf;

	tmpSurf = SDL_LoadBMP("bitmaps/background.bmp");
	bgSurf = SDL_DisplayFormat(tmpSurf);
	SDL_FreeSurface(tmpSurf);

	tmpSurf = SDL_LoadBMP("bitmaps/digits.bmp");
	digitSurf = SDL_DisplayFormat(tmpSurf);
	SDL_FreeSurface(tmpSurf);

	tmpSurf = SDL_LoadBMP("bitmaps/text_pause.bmp");
	SDL_SetColorKey(tmpSurf, SDL_SRCCOLORKEY, SDL_MapRGB(tmpSurf->format, 0, 0, 0));
	txtPauseSurf = SDL_DisplayFormat(tmpSurf);
	SDL_FreeSurface(tmpSurf);

	tmpSurf = SDL_LoadBMP("bitmaps/text_presskey.bmp");
	SDL_SetColorKey(tmpSurf, SDL_SRCCOLORKEY, SDL_MapRGB(tmpSurf->format, 0, 0, 0));
	txtPresskeySurf = SDL_DisplayFormat(tmpSurf);
	SDL_FreeSurface(tmpSurf);


	/* Create targets */
	/* Target::Target(posX, posY, posW, posH, preX, preY, preW, preH, type) */
	targets[0] = new Target(294, 186, 167, 22, 254, 186, 40, 22, SPRITETYPE_RED);
	targets[1] = new Target(423, 160, 22, 167, 423, 327, 22, 40, SPRITETYPE_GREEN);
	targets[2] = new Target(490, 255, 24, 24, 484, 249, 36, 36, SPRITETYPE_BLUE);
	numtargets = 3;


	/* Create sprites */
	for(i=0; i<MAXNUMSPRITES; i++) {
		
		SDL_Surface* srf1;
		SDL_Surface* srf2;
		SDL_Surface* srf3;
		SDL_Surface* tmpSrf1;
		SDL_Surface* tmpSrf2;
		SDL_Surface* tmpSrf3;
		float spd;
		int typ;

		if(i<4) {
			tmpSrf1 = SDL_LoadBMP("bitmaps/d1.bmp");
			tmpSrf2 = SDL_LoadBMP("bitmaps/d2.bmp");
			tmpSrf3 = SDL_LoadBMP("bitmaps/d4.bmp");
			spd = SPEED_D;
			typ = SPRITETYPE_BLUE;
		}
		else if(4<=i && i<8) {
			tmpSrf1 = SDL_LoadBMP("bitmaps/c1.bmp");
			tmpSrf2 = SDL_LoadBMP("bitmaps/c2.bmp");
			tmpSrf3 = SDL_LoadBMP("bitmaps/c4.bmp");
			spd = SPEED_C;
			typ = SPRITETYPE_GREEN;
		}
		else if(8<=i && i<14) {
			tmpSrf1 = SDL_LoadBMP("bitmaps/b1.bmp");
			tmpSrf2 = SDL_LoadBMP("bitmaps/b2.bmp");
			tmpSrf3 = SDL_LoadBMP("bitmaps/b4.bmp");
			spd = SPEED_B;
			typ = SPRITETYPE_RED;
		}
		else {
			tmpSrf1 = SDL_LoadBMP("bitmaps/a1.bmp");
			tmpSrf2 = SDL_LoadBMP("bitmaps/a2.bmp");
			tmpSrf3 = SDL_LoadBMP("bitmaps/a4.bmp");
			spd = SPEED_A;
			typ = SPRITETYPE_RED;
		}

		SDL_SetColorKey(tmpSrf1, SDL_SRCCOLORKEY, SDL_MapRGB(tmpSrf1->format, 0, 0, 0));
		SDL_SetColorKey(tmpSrf2, SDL_SRCCOLORKEY, SDL_MapRGB(tmpSrf2->format, 0, 0, 0));
		SDL_SetColorKey(tmpSrf3, SDL_SRCCOLORKEY, SDL_MapRGB(tmpSrf3->format, 0, 0, 0));
		srf1 = SDL_DisplayFormat(tmpSrf1);
		srf2 = SDL_DisplayFormat(tmpSrf2);
		srf3 = SDL_DisplayFormat(tmpSrf3);
		SDL_FreeSurface(tmpSrf1);
		SDL_FreeSurface(tmpSrf2);
		SDL_FreeSurface(tmpSrf3);

		sprites[i] = new FlyObj(srf1, srf2, srf3, spd, typ);
		numsprites++;
	}


	/* Read highscore */
	readHighscore();


	/* Game loop */
	bool done = false;
	bool pause = false;
	while(!done) {
		
		addNewSprite();
		
		/* Poll event loop */
		while(SDL_PollEvent(&event)) {

			switch(event.type) {

				case SDL_MOUSEBUTTONDOWN:
					if(!pause) {
						pause = !(handleMouseInput(&event)); // game over --> pause
					}
					break;

				case SDL_KEYDOWN:
					if(event.key.keysym.sym == SDLK_ESCAPE) {
						done = true;
					}
					else {
						if(!pause) {
							pause = true;
							pauseScreen();
						}
						else {
							pause = false;
						}
					}
					break;

				case SDL_QUIT:
					done = true;
					break;

				default:
					break;
			}

		} // poll event loop


		/* Move sprites, check for collisions, ... */
		if(!pause) {
			pause = !(doGameTickTasks()); // game over --> pause
		}

	} // game loop

	SDL_Quit();
	return(0);

}

