/* 
 * File   : pathsprite.hpp
 * Version: 2009-10-10
 *
 * Copyright (C) 2010 Jochen Peters
 *
 * dev[at]emirp.net
 * http://www.emirp.net/
 * 
 * This file is part of Approach Control.
 *
 * Approach Control is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Approach Control is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Approach Control. If not, see <http://www.gnu.org/licenses/>.
 *
 */


#ifndef SPRITE_HPP
#define SPRITE_HPP

#include <SDL.h>


#define PMAXLENGTH 1024
#define MAXSPEED 40

/* PathSprite status */
#define SPRITESTAT_IDLE 0
#define SPRITESTAT_WAITING 1
#define SPRITESTAT_FREERUNNING 2
#define SPRITESTAT_ONPATH 3
#define SPRITESTAT_ONAPPROACH 4
#define SPRITESTAT_FINALAPPROACH 5
#define SPRITESTAT_LANDED 6
#define SPRITESTAT_COLLISION 9

/* PathSprite types */
#define SPRITETYPE_RED 1
#define SPRITETYPE_GREEN 2
#define SPRITETYPE_BLUE 3


class Target;


typedef struct {
	int x;
	int y;
} Coord;


class PathSprite {
	
	public:
		PathSprite(SDL_Surface*, int, int, int, int, float, int);
		void setSurface(SDL_Surface*);
		void draw(SDL_Surface*);
		void undraw(SDL_Surface*, Uint32);
		void show(SDL_Surface*);
		void remove(SDL_Surface*, Uint32);
		void move(SDL_Surface*, Uint32);
		void setPos(int, int);
		void setVelocity(int, int);
		SDL_Rect* getPos();
		int getPosX();
		int getPosY();
		int getCPosX();
		int getCPosY();
		int getWidth();
		int getHeight();
		int getType();
		bool isVisible();
		void clearPath();
		int addToPath(int, int);
		int addToPath(Coord*);
		void drawPathRGBA(SDL_Surface*, Uint8, Uint8, Uint8, Uint8);
		bool collisionCheck(int, int);
		bool collisionCheckInside(PathSprite*);
		bool collisionCheckOutline(PathSprite*);
		bool collisionCheckInside(SDL_Rect*);
		bool collisionCheckOutline(SDL_Rect*);
		bool checkTarget(Target*);
		bool checkTarget(Target**, int);
		bool checkPath(Target*);
		bool checkPath(Target**, int);

		int status;

	protected:
		SDL_Surface* spritesurf;
		SDL_Rect pos;
		Coord centerpos;
		Coord vel;
		unsigned int speed;
		unsigned int movcount;
		Coord path[PMAXLENGTH];
		unsigned int plength;
		unsigned int pindex;
		bool visible;
		int type;

		void calcVelocity();

};


class Target {
	
	public:
		SDL_Rect pos;
		SDL_Rect pre;
		int type;

		Target(int, int, int, int, int, int, int, int, int);
		void draw(SDL_Surface*);
		bool inTarget(int, int);
		bool inPreTarget(int, int);
		int getType();

};


class Marker {
	
	public:
		Marker(PathSprite*, unsigned int);
		void draw(SDL_Surface*, Uint32);
		bool startSprite(SDL_Surface*);

	protected:
		PathSprite* sprite;
		SDL_Rect pos;
		unsigned int countdown;


};

#endif


